@@ -3,7 +3,7 @@
  Midi Player using the WildMidi Midi Processing Library
 
  Copyright (C) Chris Ison  2001-2011
- Copyright (C) Bret Curtis 2013-2014
+ Copyright (C) Bret Curtis 2013-2016
 
  This file is part of WildMIDI.
 
@@ -26,16 +26,16 @@
 
 #include "config.h"
 
-#include <sys/types.h>
 #include <limits.h>
 #include <errno.h>
-#include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/stat.h>
 
 #if defined(__DJGPP__)
+#include <sys/types.h>
+#include <fcntl.h>
+#include <sys/stat.h>
 #include "getopt_long.h"
 #include <conio.h>
 #define getopt dj_getopt /* hack */
@@ -46,29 +46,73 @@
 #ifdef AUDIODRV_DOSSB
 #include "dossb.h"
 #endif
-#endif
 
-#if (defined _WIN32) || (defined __CYGWIN__)
+#elif (defined _WIN32) || (defined __CYGWIN__)
+#include <sys/types.h>
+#include <fcntl.h>
+#include <sys/stat.h>
 #include <conio.h>
 #include <windows.h>
 #include <mmsystem.h>
 #define msleep(s) Sleep((s))
 #include <io.h>
-#undef close
-#define close _close
-#undef open
-#define open _open
-#undef read
-#define read _read
-#undef write
-#define write _write
-#undef lseek
-#define lseek _lseek
 #include "getopt_long.h"
+#ifdef __WATCOMC__
+#define _putch putch
 #endif
 
-#if !defined(_WIN32) && !defined(__DJGPP__) /* unix build */
+#elif defined(__OS2__) || defined(__EMX__)
+#define INCL_DOS
+#define INCL_DOSERRORS
+#define INCL_OS2MM
+#ifdef __EMX__
+#define INCL_KBD
+#define INCL_VIO
+#endif
+#include <os2.h>
+#include <os2me.h>
+#include <conio.h>
+#define msleep(s) DosSleep((s))
+#include <fcntl.h>
+#include <io.h>
+#include "getopt_long.h"
+#ifdef __EMX__
+#include <sys/types.h> /* for off_t typedef */
+int putch (int c) {
+	char ch = c;
+	VioWrtTTY(&ch, 1, 0);
+	return c;
+}
+int kbhit (void) {
+	KBDKEYINFO k;
+	if (KbdPeek(&k, 0))
+		return 0;
+	return (k.fbStatus & KBDTRF_FINAL_CHAR_IN);
+}
+#endif
+
+#elif defined(WILDMIDI_AMIGA)
+extern void amiga_sysinit (void);
+extern int amiga_usleep(unsigned long millisec);
+#define msleep(s) amiga_usleep((s)*1000)
+extern int amiga_getch (unsigned char *ch);
+#include <proto/exec.h>
+#include <proto/dos.h>
+#include "getopt_long.h"
+#ifdef AUDIODRV_AHI
+#include <devices/ahi.h>
+#ifdef __amigaos4__
+#define SHAREDMEMFLAG MEMF_SHARED
+#else
+#define SHAREDMEMFLAG MEMF_PUBLIC
+#endif
+#endif
+
+#else /* unix build */
 static int msleep(unsigned long millisec);
+#include <sys/types.h>
+#include <fcntl.h>
+#include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <unistd.h>
 #include <getopt.h>
@@ -212,7 +256,108 @@ static int (*send_output)(char * output_data, int output_size);
 static void (*close_output)(void);
 static void (*pause_output)(void);
 static void (*resume_output)(void);
+
+#define wmidi_geterrno() errno /* generic case */
+#if defined(_WIN32)
 static int audio_fd = -1;
+#define WM_IS_BADF(_fd) ((_fd)<0)
+#define WM_BADF -1
+static inline int wmidi_open_write (const char *path) {
+	return _open(path, (O_RDWR | O_CREAT | O_TRUNC | O_BINARY), 0664);
+}
+static inline void wmidi_close (int fd) {
+	_close(fd);
+}
+static inline long wmidi_seekset (int fd, long ofs) {
+	return _lseek(fd, ofs, SEEK_SET);
+}
+static inline int wmidi_write (int fd, const void *buf, size_t size) {
+	return _write(fd, buf, size);
+}
+
+#elif defined(__DJGPP__)
+static int audio_fd = -1;
+#define WM_IS_BADF(_fd) ((_fd)<0)
+#define WM_BADF -1
+static inline int wmidi_open_write (const char *path) {
+	return open(path, (O_RDWR | O_CREAT | O_TRUNC | O_BINARY), 0664);
+}
+static inline void wmidi_close (int fd) {
+	close(fd);
+}
+static inline off_t wmidi_seekset (int fd, off_t ofs) {
+	return lseek(fd, ofs, SEEK_SET);
+}
+static inline int wmidi_write (int fd, const void *buf, size_t size) {
+	return write(fd, buf, size);
+}
+
+#elif defined(__OS2__) || defined(__EMX__)
+static int audio_fd = -1;
+#define WM_IS_BADF(_fd) ((_fd)<0)
+#define WM_BADF -1
+static inline int wmidi_open_write (const char *path) {
+	return open(path, (O_RDWR | O_CREAT | O_TRUNC | O_BINARY), 0664);
+}
+static inline void wmidi_close (int fd) {
+	close(fd);
+}
+static inline off_t wmidi_seekset (int fd, off_t ofs) {
+	return lseek(fd, ofs, SEEK_SET);
+}
+static inline int wmidi_write (int fd, const void *buf, size_t size) {
+	return write(fd, buf, size);
+}
+
+#elif defined(WILDMIDI_AMIGA)
+static BPTR audio_fd = 0;
+#define WM_IS_BADF(_fd) ((_fd)==0)
+#define WM_BADF 0
+#undef wmidi_geterrno
+static int wmidi_geterrno (void) {
+	switch (IoErr()) {
+	case ERROR_OBJECT_NOT_FOUND: return ENOENT;
+	case ERROR_DISK_FULL: return ENOSPC;
+	}
+	return EIO; /* better ?? */
+}
+static inline BPTR wmidi_open_write (const char *path) {
+	return Open((const STRPTR) path, MODE_NEWFILE);
+}
+static inline LONG wmidi_close (BPTR fd) {
+	return Close(fd);
+}
+static inline LONG wmidi_seekset (BPTR fd, LONG ofs) {
+	return Seek(fd, ofs, OFFSET_BEGINNING);
+}
+static LONG wmidi_write (BPTR fd, /*const*/ void *buf, LONG size) {
+	LONG written = 0, result;
+	unsigned char *p = (unsigned char *)buf;
+	while (written < size) {
+		result = Write(fd, p + written, size - written);
+		if (result < 0) return result;
+		written += result;
+	}
+	return written;
+}
+
+#else /* common posix case */
+static int audio_fd = -1;
+#define WM_IS_BADF(_fd) ((_fd)<0)
+#define WM_BADF -1
+static inline int wmidi_open_write (const char *path) {
+	return open(path, (O_RDWR | O_CREAT | O_TRUNC), (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH));
+}
+static inline int wmidi_close (int fd) {
+	return close(fd);
+}
+static inline off_t wmidi_seekset (int fd, off_t ofs) {
+	return lseek(fd, ofs, SEEK_SET);
+}
+static inline ssize_t wmidi_write (int fd, const void *buf, size_t size) {
+	return write(fd, buf, size);
+}
+#endif
 
 static void pause_output_nop(void) {
 }
@@ -253,14 +398,10 @@ static int open_wav_output(void) {
 
 	if (wav_file[0] == '\0')
 		return (-1);
-#if defined(_WIN32) || defined(__DJGPP__)
-	audio_fd = open(wav_file, (O_RDWR | O_CREAT | O_TRUNC | O_BINARY), 0664);
-#else
-	audio_fd = open(wav_file, (O_RDWR | O_CREAT | O_TRUNC),
-			(S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH));
-#endif
-	if (audio_fd < 0) {
-		fprintf(stderr, "Error: unable to open file for writing (%s)\r\n", strerror(errno));
+
+	audio_fd = wmidi_open_write(wav_file);
+	if (WM_IS_BADF(audio_fd)) {
+		fprintf(stderr, "Error: unable to open file for writing (%s)\r\n", strerror(wmidi_geterrno()));
 		return (-1);
 	} else {
 		unsigned long int bytes_per_sec;
@@ -275,10 +416,10 @@ static int open_wav_output(void) {
 		wav_hdr[31] = (bytes_per_sec >> 24) & 0xFF;
 	}
 
-	if (write(audio_fd, &wav_hdr, 44) < 0) {
-		fprintf(stderr, "ERROR: failed writing wav header (%s)\r\n", strerror(errno));
-		close(audio_fd);
-		audio_fd = -1;
+	if (wmidi_write(audio_fd, wav_hdr, 44) < 0) {
+		fprintf(stderr, "ERROR: failed writing wav header (%s)\r\n", strerror(wmidi_geterrno()));
+		wmidi_close(audio_fd);
+		audio_fd = WM_BADF;
 		return (-1);
 	}
 
@@ -292,10 +433,10 @@ static int open_wav_output(void) {
 
 static int write_wav_output(char * output_data, int output_size) {
 /* the library specifically outputs LE data - no need swapping. */
-	if (write(audio_fd, output_data, output_size) < 0) {
-		fprintf(stderr, "\nERROR: failed writing wav (%s)\r\n", strerror(errno));
-		close(audio_fd);
-		audio_fd = -1;
+	if (wmidi_write(audio_fd, output_data, output_size) < 0) {
+		fprintf(stderr, "\nERROR: failed writing wav (%s)\r\n", strerror(wmidi_geterrno()));
+		wmidi_close(audio_fd);
+		audio_fd = WM_BADF;
 		return (-1);
 	}
 
@@ -305,7 +446,7 @@ static int write_wav_output(char * output_data, int output_size) {
 
 static void close_wav_output(void) {
 	unsigned char wav_count[4];
-	if (audio_fd < 0)
+	if (WM_IS_BADF(audio_fd))
 		return;
 
 	printf("Finishing and closing wav output\r");
@@ -313,9 +454,9 @@ static void close_wav_output(void) {
 	wav_count[1] = (wav_size >> 8) & 0xFF;
 	wav_count[2] = (wav_size >> 16) & 0xFF;
 	wav_count[3] = (wav_size >> 24) & 0xFF;
-	lseek(audio_fd, 40, SEEK_SET);
-	if (write(audio_fd, &wav_count, 4) < 0) {
-		fprintf(stderr, "\nERROR: failed writing wav (%s)\r\n", strerror(errno));
+	wmidi_seekset(audio_fd, 40);
+	if (wmidi_write(audio_fd, wav_count, 4) < 0) {
+		fprintf(stderr, "\nERROR: failed writing wav (%s)\r\n", strerror(wmidi_geterrno()));
 		goto end;
 	}
 
@@ -324,16 +465,16 @@ static void close_wav_output(void) {
 	wav_count[1] = (wav_size >> 8) & 0xFF;
 	wav_count[2] = (wav_size >> 16) & 0xFF;
 	wav_count[3] = (wav_size >> 24) & 0xFF;
-	lseek(audio_fd, 4, SEEK_SET);
-	if (write(audio_fd, &wav_count, 4) < 0) {
-		fprintf(stderr, "\nERROR: failed writing wav (%s)\r\n", strerror(errno));
+	wmidi_seekset(audio_fd, 4);
+	if (wmidi_write(audio_fd, wav_count, 4) < 0) {
+		fprintf(stderr, "\nERROR: failed writing wav (%s)\r\n", strerror(wmidi_geterrno()));
 		goto end;
 	}
 
 end:	printf("\n");
-	if (audio_fd >= 0)
-		close(audio_fd);
-	audio_fd = -1;
+	if (!WM_IS_BADF(audio_fd))
+		wmidi_close(audio_fd);
+	audio_fd = WM_BADF;
 }
 
 #if (defined _WIN32) || (defined __CYGWIN__)
@@ -474,6 +615,182 @@ close_mm_output (void) {
 	mm_blocks = NULL;
 }
 
+#elif (defined(__OS2__) || defined(__EMX__)) && defined(AUDIODRV_OS2DART)
+/* based on Dart code originally written by Kevin Langman for XMP */
+
+#define open_audio_output open_dart_output
+static int write_dart_output (char *output_data, int output_size);
+static void close_dart_output (void);
+
+#define BUFFERCOUNT 4
+
+static MCI_MIX_BUFFER MixBuffers[BUFFERCOUNT];
+static MCI_MIXSETUP_PARMS MixSetupParms;
+static MCI_BUFFER_PARMS BufferParms;
+static MCI_GENERIC_PARMS GenericParms;
+
+static ULONG DeviceID = 0;
+static ULONG bsize = 16;
+static short next = 2;
+static short ready = 1;
+
+static HMTX dart_mutex;
+
+/* Buffer update thread (created and called by DART) */
+static LONG APIENTRY OS2_Dart_UpdateBuffers
+	(ULONG ulStatus, PMCI_MIX_BUFFER pBuffer, ULONG ulFlags) {
+
+	(void) pBuffer;/* unused param */
+
+	if ((ulFlags == MIX_WRITE_COMPLETE) ||
+		((ulFlags == (MIX_WRITE_COMPLETE | MIX_STREAM_ERROR)) &&
+		 (ulStatus == ERROR_DEVICE_UNDERRUN))) {
+		DosRequestMutexSem(dart_mutex, SEM_INDEFINITE_WAIT);
+		ready++;
+		DosReleaseMutexSem(dart_mutex);
+	}
+	return (TRUE);
+}
+
+static int
+open_dart_output(void) {
+	int i;
+	MCI_AMP_OPEN_PARMS AmpOpenParms;
+
+	if (DosCreateMutexSem(NULL, &dart_mutex, 0, 0) != NO_ERROR) {
+		fprintf(stderr, "Failed creating a MutexSem.\r\n");
+		return (-1);
+	}
+
+	/* compute a size for circa 1/4" of playback. */
+	bsize = rate >> 2;
+	bsize <<= 1; /* stereo */
+	bsize <<= 1; /* 16 bit */
+	for (i = 15; i >= 12; i--) {
+		if (bsize & (1 << i))
+			break;
+	}
+	bsize = (1 << i);
+	/* make sure buffer is not greater than 64 Kb: DART can't handle it. */
+	if (bsize > 65536)
+		bsize = 65536;
+
+	MixBuffers[0].pBuffer = NULL; /* marker */
+	memset(&GenericParms, 0, sizeof(MCI_GENERIC_PARMS));
+
+	/* open AMP device */
+	memset(&AmpOpenParms, 0, sizeof(MCI_AMP_OPEN_PARMS));
+	AmpOpenParms.usDeviceID = 0;
+
+	AmpOpenParms.pszDeviceType =
+		(PSZ) MAKEULONG(MCI_DEVTYPE_AUDIO_AMPMIX, 0); /* 0: default waveaudio device */
+
+	if(mciSendCommand(0, MCI_OPEN, MCI_WAIT|MCI_OPEN_TYPE_ID|MCI_OPEN_SHAREABLE,
+			  (PVOID) &AmpOpenParms, 0) != MCIERR_SUCCESS) {
+		fprintf(stderr, "Failed opening DART audio device\r\n");
+		return (-1);
+	}
+
+	DeviceID = AmpOpenParms.usDeviceID;
+
+	/* setup playback parameters */
+	memset(&MixSetupParms, 0, sizeof(MCI_MIXSETUP_PARMS));
+
+	MixSetupParms.ulBitsPerSample = 16;
+	MixSetupParms.ulFormatTag = MCI_WAVE_FORMAT_PCM;
+	MixSetupParms.ulSamplesPerSec = rate;
+	MixSetupParms.ulChannels = 2;
+	MixSetupParms.ulFormatMode = MCI_PLAY;
+	MixSetupParms.ulDeviceType = MCI_DEVTYPE_WAVEFORM_AUDIO;
+	MixSetupParms.pmixEvent = OS2_Dart_UpdateBuffers;
+
+	if (mciSendCommand(DeviceID, MCI_MIXSETUP,
+			   MCI_WAIT | MCI_MIXSETUP_INIT,
+			   (PVOID) & MixSetupParms, 0) != MCIERR_SUCCESS) {
+
+		mciSendCommand(DeviceID, MCI_CLOSE, MCI_WAIT,
+			       (PVOID) & GenericParms, 0);
+		fprintf(stderr, "Failed DART mixer setup\r\n");
+		return (-1);
+	}
+
+	/*bsize = MixSetupParms.ulBufferSize;*/
+	/*printf("Dart Buffer Size = %lu\n", bsize);*/
+
+	BufferParms.ulNumBuffers = BUFFERCOUNT;
+	BufferParms.ulBufferSize = bsize;
+	BufferParms.pBufList = MixBuffers;
+
+	if (mciSendCommand(DeviceID, MCI_BUFFER,
+			   MCI_WAIT | MCI_ALLOCATE_MEMORY,
+			   (PVOID) & BufferParms, 0) != MCIERR_SUCCESS) {
+		fprintf(stderr, "DART Memory allocation error\r\n");
+		mciSendCommand(DeviceID, MCI_CLOSE, MCI_WAIT,
+			       (PVOID) & GenericParms, 0);
+		return (-1);
+	}
+
+	for (i = 0; i < BUFFERCOUNT; i++) {
+		MixBuffers[i].ulBufferLength = bsize;
+	}
+
+	/* Start Playback */
+	memset(MixBuffers[0].pBuffer, /*32767 */ 0, bsize);
+	memset(MixBuffers[1].pBuffer, /*32767 */ 0, bsize);
+	MixSetupParms.pmixWrite(MixSetupParms.ulMixHandle, MixBuffers, 2);
+
+	send_output = write_dart_output;
+	close_output = close_dart_output;
+	pause_output = pause_output_nop;
+	resume_output = resume_output_nop;
+
+	return (0);
+}
+
+static int
+write_dart_output (char *output_data, int output_size) {
+	static int idx = 0;
+
+	if (idx + output_size > bsize) {
+		do {
+			DosRequestMutexSem(dart_mutex, SEM_INDEFINITE_WAIT);
+			if (ready != 0) {
+				DosReleaseMutexSem(dart_mutex);
+				break;
+			}
+			DosReleaseMutexSem(dart_mutex);
+			DosSleep(20);
+		} while (TRUE);
+
+		MixBuffers[next].ulBufferLength = idx;
+		MixSetupParms.pmixWrite(MixSetupParms.ulMixHandle, &(MixBuffers[next]), 1);
+		ready--;
+		next++;
+		idx = 0;
+		if (next == BUFFERCOUNT) {
+			next = 0;
+		}
+	}
+	memcpy(&((char *)MixBuffers[next].pBuffer)[idx], output_data, output_size);
+	idx += output_size;
+	return (0);
+}
+
+static void
+close_dart_output (void) {
+	printf("Shutting down sound output\r\n");
+	if (MixBuffers[0].pBuffer) {
+		mciSendCommand(DeviceID, MCI_BUFFER,
+			       MCI_WAIT | MCI_DEALLOCATE_MEMORY, &BufferParms, 0);
+		MixBuffers[0].pBuffer = NULL;
+	}
+	if (DeviceID) {
+		mciSendCommand(DeviceID, MCI_CLOSE, MCI_WAIT,
+			       (PVOID) &GenericParms, 0);
+		DeviceID = 0;
+	}
+}
+
 #elif defined(__DJGPP__) && defined(AUDIODRV_DOSSB)
 /* SoundBlaster/Pro/16/AWE32 driver for DOS -- adapted from
  * libMikMod,  written by Andrew Zabolotny <bit@eltech.ru>,
@@ -643,6 +960,120 @@ static int open_sb_output(void)
 	return 0;
 }
 
+#elif defined(WILDMIDI_AMIGA) && defined(AUDIODRV_AHI)
+
+/* Driver for output to native Amiga AHI device:
+ * Written by Szilárd Biró <col.lawrence@gmail.com>, loosely based
+ * on an old AOS4 version by Fredrik Wikstrom <fredrik@a500.org>
+ */
+
+#define BUFFERSIZE (4 << 10)
+
+static struct MsgPort *AHImp = NULL;
+static struct AHIRequest *AHIReq[2] = { NULL, NULL };
+static int active = 0;
+static char *AHIBuf[2] = { NULL, NULL };
+
+#define open_audio_output open_ahi_output
+static int write_ahi_output(char * output_data, int output_size);
+static void close_ahi_output(void);
+
+static int open_ahi_output(void) {
+	AHImp = CreateMsgPort();
+	if (AHImp) {
+		AHIReq[0] = (struct AHIRequest *)CreateIORequest(AHImp, sizeof(struct AHIRequest));
+		if (AHIReq[0]) {
+			AHIReq[0]->ahir_Version = 4;
+			AHIReq[1] = AllocVec(sizeof(struct AHIRequest), SHAREDMEMFLAG);
+			if (AHIReq[1]) {
+				if (!OpenDevice(AHINAME, AHI_DEFAULT_UNIT, (struct IORequest *)AHIReq[0], 0)) {
+					/*AHIReq[0]->ahir_Std.io_Message.mn_Node.ln_Pri = 0;*/
+					AHIReq[0]->ahir_Std.io_Command = CMD_WRITE;
+					AHIReq[0]->ahir_Std.io_Data = NULL;
+					AHIReq[0]->ahir_Std.io_Offset = 0;
+					AHIReq[0]->ahir_Frequency = rate;
+					AHIReq[0]->ahir_Type = AHIST_S16S;/* 16 bit stereo */
+					AHIReq[0]->ahir_Volume = 0x10000;
+					AHIReq[0]->ahir_Position = 0x8000;
+					CopyMem(AHIReq[0], AHIReq[1], sizeof(struct AHIRequest));
+
+					AHIBuf[0] = AllocVec(BUFFERSIZE, SHAREDMEMFLAG | MEMF_CLEAR);
+					if (AHIBuf[0]) {
+						AHIBuf[1] = AllocVec(BUFFERSIZE, SHAREDMEMFLAG | MEMF_CLEAR);
+						if (AHIBuf[1]) {
+							send_output = write_ahi_output;
+							close_output = close_ahi_output;
+							pause_output = pause_output_nop;
+							resume_output = resume_output_nop;
+							return (0);
+						}
+					}
+				}
+			}
+		}
+	}
+
+	close_ahi_output();
+	fprintf(stderr, "ERROR: Unable to open AHI output\r\n");
+	return (-1);
+}
+
+static int write_ahi_output(char * output_data, int output_size) {
+	int chunk;
+	while (output_size > 0) {
+		if (AHIReq[active]->ahir_Std.io_Data) {
+			WaitIO((struct IORequest *) AHIReq[active]);
+		}
+		chunk = (output_size < BUFFERSIZE)? output_size : BUFFERSIZE;
+		memcpy(AHIBuf[active], output_data, chunk);
+		output_size -= chunk;
+		output_data += chunk;
+
+		AHIReq[active]->ahir_Std.io_Data = AHIBuf[active];
+		AHIReq[active]->ahir_Std.io_Length = chunk;
+		AHIReq[active]->ahir_Link = !CheckIO((struct IORequest *) AHIReq[active ^ 1]) ? AHIReq[active ^ 1] : NULL;
+		SendIO((struct IORequest *)AHIReq[active]);
+		active ^= 1;
+	}
+	return (0);
+}
+
+static void close_ahi_output(void) {
+	if (AHIReq[1]) {
+		AHIReq[0]->ahir_Link = NULL; /* in case we are linked to req[0] */
+		if (!CheckIO((struct IORequest *) AHIReq[1])) {
+			AbortIO((struct IORequest *) AHIReq[1]);
+			WaitIO((struct IORequest *) AHIReq[1]);
+		}
+		FreeVec(AHIReq[1]);
+		AHIReq[1] = NULL;
+	}
+	if (AHIReq[0]) {
+		if (!CheckIO((struct IORequest *) AHIReq[0])) {
+			AbortIO((struct IORequest *) AHIReq[0]);
+			WaitIO((struct IORequest *) AHIReq[0]);
+		}
+		if (AHIReq[0]->ahir_Std.io_Device) {
+			CloseDevice((struct IORequest *) AHIReq[0]);
+			AHIReq[0]->ahir_Std.io_Device = NULL;
+		}
+		DeleteIORequest((struct IORequest *) AHIReq[0]);
+		AHIReq[0] = NULL;
+	}
+	if (AHImp) {
+		DeleteMsgPort(AHImp);
+		AHImp = NULL;
+	}
+	if (AHIBuf[0]) {
+		FreeVec(AHIBuf[0]);
+		AHIBuf[0] = NULL;
+	}
+	if (AHIBuf[1]) {
+		FreeVec(AHIBuf[1]);
+		AHIBuf[1] = NULL;
+	}
+}
+
 #else
 #ifdef AUDIODRV_ALSA
 
@@ -679,8 +1110,7 @@ static int open_alsa_output(void) {
 		goto fail;
 	}
 
-	if ((err = snd_pcm_hw_params_set_access(pcm, hw,
-			SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
+	if ((err = snd_pcm_hw_params_set_access(pcm, hw, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
 		fprintf(stderr, "Cannot set access mode: %s.\r\n", snd_strerror(err));
 		goto fail;
 	}
@@ -708,14 +1138,12 @@ static int open_alsa_output(void) {
 	alsa_buffer_time = 500000;
 	alsa_period_time = 50000;
 
-	if ((err = snd_pcm_hw_params_set_buffer_time_near(pcm, hw,
-			&alsa_buffer_time, 0)) < 0) {
+	if ((err = snd_pcm_hw_params_set_buffer_time_near(pcm, hw, &alsa_buffer_time, 0)) < 0) {
 		fprintf(stderr, "Set buffer time failed: %s.\r\n", snd_strerror(err));
 		goto fail;
 	}
 
-	if ((err = snd_pcm_hw_params_set_period_time_near(pcm, hw,
-			&alsa_period_time, 0)) < 0) {
+	if ((err = snd_pcm_hw_params_set_period_time_near(pcm, hw, &alsa_period_time, 0)) < 0) {
 		fprintf(stderr, "Set period time failed: %s.\r\n", snd_strerror(err));
 		goto fail;
 	}
@@ -808,7 +1236,7 @@ static int open_oss_output(void) {
 
 	if ((audio_fd = open(pcmname, O_WRONLY)) < 0) {
 		fprintf(stderr, "ERROR: Unable to open dsp (%s)\r\n", strerror(errno));
-		return -1;
+		return (-1);
 	}
 	if (ioctl(audio_fd, SNDCTL_DSP_RESET, 0) < 0) {
 		fprintf(stderr, "ERROR: Unable to reset dsp\r\n");
@@ -850,7 +1278,7 @@ static int open_oss_output(void) {
 	return (0);
 
 fail:	close_oss_output();
-	return -1;
+	return (-1);
 }
 
 static int write_oss_output(char * output_data, int output_size) {
@@ -863,7 +1291,7 @@ static int write_oss_output(char * output_data, int output_size) {
 		} else {
 			fprintf(stderr, "\nOSS: write failure to dsp: %s.\r\n",
 				strerror(errno));
-			return -1;
+			return (-1);
 		}
 	}
 	return (0);
@@ -1004,6 +1432,7 @@ static int open_openal_output(void) {
 
 #define open_audio_output open_noaudio_output
 static int open_noaudio_output(void) {
+	fprintf(stderr, "No audio output driver was selected at compile time.\r\n");
 	return -1;
 }
 
@@ -1052,7 +1481,7 @@ static void do_help(void) {
 
 static void do_version(void) {
 	printf("\nWildMidi %s Open Source Midi Sequencer\n", PACKAGE_VERSION);
-	printf("Copyright (C) WildMIDI Developers 2001-2014\n\n");
+	printf("Copyright (C) WildMIDI Developers 2001-2016\n\n");
 	printf("WildMidi comes with ABSOLUTELY NO WARRANTY\n");
 	printf("This is free software, and you are welcome to redistribute it under\n");
 	printf("the terms and conditions of the GNU General Public License version 3.\n");
@@ -1083,8 +1512,8 @@ int main(int argc, char **argv) {
 	char modes[4];
 	unsigned long int count_diff;
 	unsigned char ch;
-	unsigned char test_midi = 0;
-	unsigned char test_count = 0;
+	int test_midi = 0;
+	int test_count = 0;
 	unsigned char *test_data;
 	unsigned char test_bank = 0;
 	unsigned char test_patch = 0;
@@ -1214,6 +1643,9 @@ int main(int argc, char **argv) {
 		}
 
 		wm_inittty();
+#ifdef WILDMIDI_AMIGA
+		amiga_sysinit();
+#endif
 
 		WildMidi_MasterVolume(master_volume);
 
@@ -1272,11 +1704,13 @@ int main(int argc, char **argv) {
 					ch = _getch();
 					_putch(ch);
 				}
-#elif defined(__DJGPP__)
+#elif defined(__DJGPP__) || defined(__OS2__) || defined(__EMX__)
 				if (kbhit()) {
 					ch = getch();
 					putch(ch);
 				}
+#elif defined(WILDMIDI_AMIGA)
+				amiga_getch (&ch);
 #else
 				if (read(STDIN_FILENO, &ch, 1) != 1)
 					ch = 0;
@@ -1363,7 +1797,7 @@ int main(int argc, char **argv) {
 					pro_secs = (wm_info->current_sample % (rate * 60)) / rate;
 					fprintf(stderr,
 						"        [Approx %2lum %2lus Total] [%s] [%3i] [%2lum %2lus Processed] [%2lu%%] 0  \r",
-						apr_mins, apr_secs, modes, master_volume, pro_mins,
+						apr_mins, apr_secs, modes, (int)master_volume, pro_mins,
 						pro_secs, perc_play);
 
 					msleep(5);
@@ -1382,19 +1816,28 @@ int main(int argc, char **argv) {
 				pro_secs = (wm_info->current_sample % (rate * 60)) / rate;
 				fprintf(stderr,
 						"        [Approx %2lum %2lus Total] [%s] [%3i] [%2lum %2lus Processed] [%2lu%%] %c  \r",
-						apr_mins, apr_secs, modes, master_volume, pro_mins,
+						apr_mins, apr_secs, modes, (int)master_volume, pro_mins,
 						pro_secs, perc_play, spinner[spinpoint++ % 4]);
 
 #if defined(AUDIODRV_OPENAL) && defined(WORDS_BIGENDIAN)
 				if (context != NULL) { /* OpenAL is active, not wav writer */
-					unsigned short *swp =
-						(unsigned short *) output_buffer;
+					unsigned short *swp = (unsigned short *) output_buffer;
 					/* the library specifically outputs LE data
 					 * but OpenAL expects host-endian: do swap. */
 					for (i = 0; i < res / 2; ++i) {
 						swp[i] = (swp[i] << 8) | (swp[i] >> 8);
 					}
 				}
+#endif
+#if defined(AUDIODRV_AHI) && defined(WORDS_BIGENDIAN)
+				if (AHIBuf[0] != NULL) { /* AHI is active, not wav writer */
+					unsigned short *swp = (unsigned short *) output_buffer;
+					/* the library specifically outputs LE data
+					 * but AHI expects host-endian: do swap.  */
+					for (i = 0; i < res / 2; ++i) {
+						swp[i] = (swp[i] << 8) | (swp[i] >> 8);
+					}
+				}
 #endif
 				if (send_output(output_buffer, res) < 0) {
 				/* driver prints an error message already. */
@@ -1429,7 +1872,7 @@ end2:		close_output();
 
 /* helper / replacement functions: */
 
-#if !defined(_WIN32) && !defined(__DJGPP__)
+#if !(defined(_WIN32) || defined(__DJGPP__) || defined(WILDMIDI_AMIGA) || defined(__OS2__) || defined(__EMX__))
 static int msleep(unsigned long milisec) {
 	struct timespec req = { 0, 0 };
 	time_t sec = (int) (milisec / 1000);
